import logging
import unittest
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

from logging_utilities.context import get_logging_context
from logging_utilities.context import remove_logging_context
from logging_utilities.context import set_logging_context
from logging_utilities.context.thread_context import ThreadMappingContext
from logging_utilities.formatters.json_formatter import JsonFormatter
from logging_utilities.log_record import reset_log_record_factory
from logging_utilities.log_record import set_log_record_ignore_missing_factory

initial_factory = logging.getLogRecordFactory()


def create_dummy_log(factory):
    return factory(__name__, logging.DEBUG, __file__, 0, '', [], None)


class ThreadContextTest(unittest.TestCase):
    # pylint: disable=invalid-name

    def test_thread_context_empty(self):
        ctx = ThreadMappingContext()
        self.assertEqual(len(ctx), 0)
        self.assertEqual(ctx, {})

    def test_thread_context_init(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1})
        self.assertNotEqual(len(ctx), 0)
        self.assertEqual(ctx, {'a': 1})

        ctx.init({'c': 3})
        self.assertEqual(ctx, {'c': 3})

    def test_thread_context_init_invalid(self):
        ctx = ThreadMappingContext()
        self.assertRaises(ValueError, ctx.init, 'a string')

    def test_thread_context_len(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1})
        self.assertEqual(len(ctx), 1)

    def test_thread_context_get(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1})
        self.assertEqual(ctx.get('b', 'not found'), 'not found')
        try:
            self.assertEqual(ctx.get('b'), None)
        except KeyError:
            self.fail('get() should never raise KeyError')

        self.assertEqual(ctx.get('a'), 1)

        self.assertEqual(ctx['a'], 1)

    def test_thread_context_pop(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2})

        self.assertEqual(ctx.pop('a'), 1)
        self.assertNotIn('a', ctx)

        self.assertIn('b', ctx)

        self.assertRaises(KeyError, ctx.pop, 'c')
        self.assertIsNone(ctx.pop('c', None))
        self.assertEqual(ctx.pop('c', 'not found'), 'not found')

    def test_thread_context_set(self):
        ctx = ThreadMappingContext()

        ctx['a'] = 1
        self.assertEqual(ctx, {'a': 1})

        ctx.set('b', 2)
        self.assertEqual(ctx, {'a': 1, 'b': 2})

    def test_thread_context_del(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2, 'c': 3})

        del ctx['a']
        self.assertEqual(ctx, {'b': 2, 'c': 3})

        ctx.delete('b')
        self.assertEqual(ctx, {'c': 3})

    def test_thread_context_clear(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2, 'c': 3})

        ctx.clear()
        self.assertEqual(ctx, {})

    def test_thread_context_contains(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2})

        self.assertTrue('a' in ctx)
        self.assertFalse('c' in ctx)

    def test_thread_context_iter(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2})

        for k, v in ctx.items():
            self.assertIn(k, ['a', 'b'])
            if k == 'a':
                self.assertEqual(v, 1)
            elif k == 'b':
                self.assertEqual(v, 2)
            else:
                self.fail(f'Invalid key {k}')

        self.assertListEqual(list(ctx.keys()), ['a', 'b'])

        self.assertListEqual(list(ctx.values()), [1, 2])

    def test_thread_context_str(self):
        ctx = ThreadMappingContext()
        ctx.init({'a': 1, 'b': 2, 'c': 'my string'})
        self.assertEqual(str(ctx), "{'a': 1, 'b': 2, 'c': 'my string'}")


class LoggingContextTest(unittest.TestCase):

    def tearDown(self):
        super().tearDown()
        remove_logging_context()
        self.assertEqual(initial_factory, logging.getLogRecordFactory())

    def test_logging_context_empty(self):
        set_logging_context()
        factory_with_context = logging.getLogRecordFactory()

        # Make sure the factory has been changed
        self.assertNotEqual(initial_factory, factory_with_context)

        # Create a dummy log to check the contex
        record = create_dummy_log(factory_with_context)
        self.assertTrue(hasattr(record, 'context'))
        self.assertIn('context', record.__dict__)
        self.assertEqual({}, record.context)

    def test_logging_context_set(self):
        ctx1 = {'a': 1}
        set_logging_context(ctx1)
        record = create_dummy_log(logging.getLogRecordFactory())
        self.assertEqual(ctx1, record.context)

        ctx2 = {'a': 2}
        set_logging_context(ctx2)
        record = create_dummy_log(logging.getLogRecordFactory())
        self.assertEqual(ctx2, record.context)

    def test_logging_context_get(self):
        self.assertIsNone(get_logging_context())

        ctx1 = {'a': 1}
        set_logging_context(ctx1)

        self.assertEqual(ctx1, get_logging_context())

    def test_logging_context_modify(self):
        set_logging_context()
        factory_with_context = logging.getLogRecordFactory()

        record = create_dummy_log(factory_with_context)
        self.assertTrue(hasattr(record, 'context'))
        self.assertIn('context', record.__dict__)
        self.assertEqual(record.context, {})

        # Modify the context
        context = get_logging_context()
        context['a'] = 'added a string'
        record = create_dummy_log(factory_with_context)
        self.assertTrue(hasattr(record, 'context'))
        self.assertIn('context', record.__dict__)
        self.assertEqual(record.context, context)
        self.assertEqual(record.__dict__['context'], context)

        # Modify the context again
        context['a'] = 1
        context['b'] = 2
        self.assertEqual(record.context, context)

    def test_logging_context_set_in_thread(self):
        set_logging_context()
        contexts = {'1': {'a': 1}, '2': {'a': 2}, '3': {'b': 3}}

        def _test_logging_context_thread(ctx_id):
            set_logging_context(contexts[ctx_id])
            record = create_dummy_log(logging.getLogRecordFactory())
            self.assertTrue(hasattr(record, 'context'))
            self.assertEqual(contexts[ctx_id], record.context)
            return str(record.context)

        with ThreadPoolExecutor(max_workers=len(contexts)) as executor:
            future_to_ctx_id = {
                executor.submit(_test_logging_context_thread, ctx_id): ctx_id for ctx_id in contexts
            }
            for future in as_completed(future_to_ctx_id):
                ctx_id = future_to_ctx_id[future]
                try:
                    context = future.result()
                except Exception as exception:  # pylint: disable=broad-except
                    self.fail(f'Excpetion {exception} raised in thread')
                self.assertEqual(str(contexts[ctx_id]), context)

    def test_logging_context_with_custom_log_record(self):
        record = create_dummy_log(logging.getLogRecordFactory())
        with self.assertRaises(KeyError):
            record.__dict__['non_existant_attribute']  # pylint: disable=pointless-statement

        set_log_record_ignore_missing_factory()
        record = create_dummy_log(logging.getLogRecordFactory())
        self.assertEqual(record.__dict__['non_existant_attribute'], '')

        set_logging_context({'a': 'my-context'})
        record = create_dummy_log(logging.getLogRecordFactory())
        self.assertTrue(hasattr(record, 'context'))
        self.assertEqual(record.context, {'a': 'my-context'})
        self.assertEqual(record.__dict__['non_existant_attribute'], '')
        reset_log_record_factory()

        record = create_dummy_log(logging.getLogRecordFactory())
        self.assertFalse(hasattr(record, 'context'))

    def test_logging_context_logger_standard_fmt(self):
        set_logging_context({'a': 'my-context'})
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            logger.setLevel(logging.DEBUG)

            for handler in logger.handlers:
                formatter = logging.Formatter("%(message)s - %(context)s")
                handler.setFormatter(formatter)

            logger.debug('My message with context')
        self.assertEqual(ctx.output[0], "My message with context - {'a': 'my-context'}")

    def test_logging_context_logger_json_fmt(self):
        set_logging_context({'a': 'my-context'})
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            logger.setLevel(logging.DEBUG)

            for handler in logger.handlers:
                formatter = JsonFormatter(
                    {
                        "message": "message", "context": "context"
                    },
                    default=dict,
                )
                handler.setFormatter(formatter)

            logger.debug('My message with context')
        self.assertEqual(
            ctx.output[0], '{"message": "My message with context", "context": {"a": "my-context"}}'
        )

    def test_logging_context_logger_json_fmt_sub_element(self):
        set_logging_context({'a': 'my-context'})
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            logger.setLevel(logging.DEBUG)

            for handler in logger.handlers:
                formatter = JsonFormatter(
                    {
                        "message": "message", "context": "context.a"
                    },
                    default=dict,
                )
                handler.setFormatter(formatter)

            logger.debug('My message with context')
        self.assertEqual(
            ctx.output[0], '{"message": "My message with context", "context": "my-context"}'
        )



import threading
from collections.abc import Mapping

from .base import BaseContext


class ThreadMappingContext(BaseContext):
    '''Thread local mapping contex

    This class implements all Mapping (e.g. dictionary) functionality but on a
    thread local context.
    '''
    __marker = object()

    def __init__(self):
        self.__local = threading.local()
        self.__local.data = {}

    def __str__(self):
        return str(self.__local.data)

    def __getitem__(self, __key):
        return self.__local.data[__key]

    def __setitem__(self, __key, __value):
        self.__local.data[__key] = __value

    def __delitem__(self, __key):
        del self.__local.data[__key]

    def __len__(self):
        return len(self.__local.data)

    def __iter__(self):
        return self.__local.data.__iter__()

    def __contains__(self, __o):
        return self.__local.data.__contains__(__o)

    def init(self, data=None):
        if data is None:
            self.__local.data = {}
        else:
            if not isinstance(data, Mapping):
                raise ValueError('Data must be a Mapping sequence')
            self.__local.data = data

    def get(self, key, default=None):
        return self.__local.data.get(key, default)

    def pop(self, key, default=__marker):
        if default == self.__marker:
            return self.__local.data.pop(key)
        return self.__local.data.pop(key, default)

    def set(self, key, value):
        self.__local.data[key] = value

    def delete(self, key):
        del self.__local.data[key]

    def clear(self):
        self.__local.data = {}



import json
import logging
import re
import sys
import unittest
from collections import OrderedDict
from datetime import datetime

from nose2.tools import params

from logging_utilities.filters import ConstAttribute
from logging_utilities.formatters.json_formatter import JsonFormatter

# From python3.7, dict is ordered
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict


class BasicJsonFormatterTest(unittest.TestCase):
    maxDiff = None

    @classmethod
    def _configure_logger(
        cls,
        logger,
        fmt=None,
        add_always_extra=False,
        remove_empty=False,
        ignore_missing=False,
        style='%'
    ):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            formatter = JsonFormatter(
                fmt,
                add_always_extra=add_always_extra,
                remove_empty=remove_empty,
                ignore_missing=ignore_missing,
                style=style
            )
            handler.setFormatter(formatter)

    @classmethod
    def _configure_logger_with_const_attribute(cls, logger, remove_empty=False):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            const_attribute = ConstAttribute(application="test_application", empty_attr='')
            handler.addFilter(const_attribute)
            formatter = JsonFormatter(
                dictionary([
                    ('level', 'levelname'),
                    ('app', 'application'),
                    ('empty_attr', 'empty_attr'),
                    ('message', 'message'),
                ]),
                remove_empty=remove_empty,
                filter_attributes=['application', 'empty_attr']
            )
            handler.setFormatter(formatter)

    def test_string_fmt_param(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(logger, '{"level": "levelname", "message": "message"}')
            logger.info('Simple message')

        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message, dictionary([
                ("level", "INFO"),
                ("message", "Simple message"),
            ])
        )

    def test_simple_log(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(logger)
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
        message1 = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        message2 = json.loads(ctx.output[1], object_pairs_hook=dictionary)
        message3 = json.loads(ctx.output[2], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message1,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Simple message"),
            ])
        )
        self.assertDictEqual(
            message2,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Composed message: this is a composed message"),
            ])
        )
        self.assertDictEqual(
            message3,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Composed message with extra"),
            ])
        )

    if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
        # Python version prior to 3.8 do not raise any exception on non existing attribute
        @params(
            'non.existing', '%(non_existing)s', '%(non.existing)s', '%(non.existing)s with text'
        )
        def test_missing_attribute_raise(self, attribute):
            with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
                logger = logging.getLogger('test_formatter')
                self._configure_logger(
                    logger,
                    dictionary([
                        ('level', 'levelname'),
                        ('non-existing', attribute),
                        ('message', 'message'),
                    ]),
                )
                with self.assertRaises((ValueError, KeyError)):
                    logger.info('Simple message')
            self.assertListEqual(ctx.output, [])

    @params(
        (
            '%',
            dictionary([
                ('time', '%(asctime)s'),
                ('level', dictionary([('name', '%(levelname)s'), ('level', '%(levelno)d')])),
                ('pid_thid', '%(process)x/%(thread)x'),
                ('message', '%(message)s'),
            ])
        ),
        (
            '%',
            dictionary([
                ('time', 'asctime'),
                ('level', dictionary([('name', 'levelname'), ('level', '%(levelno)d')])),
                ('pid_thid', '%(process)x/%(thread)x'),
                ('message', 'message'),
            ])
        ),
        (
            '{',
            dictionary([
                ('time', '{asctime}'),
                ('level', dictionary([('name', '{levelname}'), ('level', '{levelno:d}')])),
                ('pid_thid', '{process:x}/{thread:x}'),
                ('message', '{message}'),
            ])
        ),
        (
            '$',
            dictionary([
                ('time', '${asctime}'),
                ('level', dictionary([('name', '${levelname}'), ('level', '${levelno}')])),
                ('pid_thid', '${process}/${thread}'),
                ('message', '${message}'),
            ])
        ),
    )
    def test_log_format_styles(self, style, fmt):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(logger, fmt=fmt, style=style)
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
        messages = [
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            json.loads(ctx.output[2], object_pairs_hook=dictionary)
        ]
        for message in messages:
            self.assertListEqual(list(message.keys()), ['time', 'level', 'pid_thid', 'message'])
            try:
                time_attr = datetime.strptime(message['time'], "%Y-%m-%d %X,%f")
            except ValueError as error:
                self.fail('Invalid time format in %s' % message['time'])
            self.assertAlmostEqual(time_attr.timestamp(), datetime.now().timestamp(), delta=10)
            self.assertListEqual(list(message['level'].keys()), ['name', 'level'])
            self.assertEqual(message['level']['level'], str(logging.INFO))
            self.assertIsNotNone(
                re.match(r'^[a-fA-F0-9]+/[a-fA-F0-9]+$', message['pid_thid']),
                msg='Invalid pid_thid value: %s' % message['pid_thid']
            )
        self.assertEqual(messages[0]['message'], "Simple message")
        self.assertEqual(messages[1]['message'], "Composed message: this is a composed message")
        self.assertEqual(messages[2]['message'], "Composed message with extra")

    def test_log_levelno(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('levelno_int', 'levelno'),
                    ('levelno_str', '%(levelno)d'),
                    ('message', '%(message)s'),
                ])
            )
            logger.info('Simple message')
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertEqual(message['levelno_int'], logging.INFO)
        self.assertEqual(message['levelno_str'], str(logging.INFO))

    def test_embedded_object(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('levelname', 'levelname'),
                    ('msg', 'message'),
                    (
                        'system',
                        dictionary([
                            ('context', dictionary([
                                ('module', 'module'),
                                ('file', 'filename'),
                            ]))
                        ])
                    ),
                ])
            )
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
        message1 = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        message2 = json.loads(ctx.output[1], object_pairs_hook=dictionary)
        message3 = json.loads(ctx.output[2], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message1,
            dictionary([
                ("levelname", "INFO"),
                ("msg", "Simple message"),
                (
                    "system",
                    dictionary([(
                        "context",
                        dictionary([("module", "test_json_formatter"),
                                    ("file", "test_json_formatter.py")])
                    )])
                ),
            ])
        )
        self.assertDictEqual(
            message2,
            dictionary([
                ("levelname", "INFO"),
                ("msg", "Composed message: this is a composed message"),
                (
                    "system",
                    dictionary([(
                        "context",
                        dictionary([("module", "test_json_formatter"),
                                    ("file", "test_json_formatter.py")])
                    )])
                ),
            ])
        )
        self.assertDictEqual(
            message3,
            dictionary([
                ("levelname", "INFO"),
                ("msg", "Composed message with extra"),
                (
                    "system",
                    dictionary([(
                        "context",
                        dictionary([("module", "test_json_formatter"),
                                    ("file", "test_json_formatter.py")])
                    )])
                ),
            ])
        )

    def test_list_embedded_object(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ("levelname", "levelname"),
                    ("name", "name"),
                    ("message", "message"),
                    (
                        "list", [
                            dictionary([("file", "filename")]),
                            dictionary([("module", "module")]),
                        ]
                    ),
                ])
            )
            logger.info('Composed message %s', 'with list')
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Composed message with list"),
                ('list', [{
                    "file": "test_json_formatter.py"
                }, {
                    "module": "test_json_formatter"
                }]),
            ])
        )

    def test_list_embedded_list(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ("levelname", "levelname"),
                    ("name", "name"),
                    ("message", "message"),
                    ("list", [["filename", "module"], ["levelname", "levelno"]]),
                ])
            )
            logger.info('Composed message %s', 'with embedded list')
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Composed message with embedded list"),
                (
                    'list', [["test_json_formatter.py", "test_json_formatter"],
                             ["INFO", logging.INFO]]
                ),
            ])
        )

    def test_extra_list(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(logger, add_always_extra=True)
            logger.info(
                'Composed message %s',
                'with extra',
                extra={'extra-list': [{
                    "extra-1": "file"
                }, {
                    "extra-2": "module"
                }]}
            )
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message,
            dictionary([
                ("levelname", "INFO"),
                ("name", "test_formatter"),
                ("message", "Composed message with extra"),
                ('extra-list', [{
                    "extra-1": "file"
                }, {
                    "extra-2": "module"
                }]),
            ])
        )

    def test_stack_info(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('levelname', 'levelname'),
                    ('name', 'name'),
                    ('message', 'message'),
                ])
            )
            logger.info('Message with stack info', stack_info=True)
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertListEqual(list(message.keys()), ['levelname', 'name', 'message', 'stack_info'])
        self.assertEqual(message['levelname'], 'INFO')
        self.assertEqual(message['name'], 'test_formatter')
        self.assertEqual(message['message'], 'Message with stack info')
        self.assertTrue(
            message['stack_info'].startswith('Stack (most recent call last):\n'),
            msg='stack_info do not start with "Stack (most recent call last):\\n"'
        )

    def test_exception(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('levelname', 'levelname'),
                    ('name', 'name'),
                    ('message', 'message'),
                    (
                        'exception_dict',
                        dictionary([("exc_info", "exc_info"), ("exc_text", "exc_text")])
                    ),
                    ('exception_list', ["exc_info", "exc_info"]),
                    ("exc_info", "exc_info"),
                    ("exc_text", "exc_text"),
                ])
            )
            try:
                raise ValueError("My value error")
            except ValueError as err:
                logger.exception('Exception message: %s', err)
        message = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        exc_text = 'Traceback (most recent call last):\n' + \
                   '    raise ValueError("My value error")\n' + \
                   'ValueError: My value error'
        expected_msg = dictionary([
            ("levelname", "ERROR"),
            ("name", "test_formatter"),
            ("message", "Exception message: My value error"),
            ("exception_dict", dictionary([("exc_info", True), ("exc_text", exc_text)])),
            ("exception_list", [True, exc_text]),
            ("exc_info", True),
            ("exc_text", exc_text),
        ])
        self.assertListEqual(list(message.keys()), list(expected_msg.keys()))
        self.assertEqual(message['levelname'], expected_msg['levelname'])
        self.assertEqual(message['name'], expected_msg['name'])
        self.assertEqual(message['message'], expected_msg['message'])
        self.assertEqual(message['exc_text'][:36], expected_msg['exc_text'][:36])
        self.assertEqual(message['exc_text'][-66:], expected_msg['exc_text'][-66:])

    def test_remove_empty(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger_with_const_attribute(logger, remove_empty=True)
            logger.info('Composed message %s', 'remove empty')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("app", "test_application"),
                ("message", "Composed message remove empty"),
            ])
        )

    def test_remove_non_existing(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ("level", "levelname"),
                    ("message", "message"),
                    ("non-existing-attribute", "%(nonExistingAttribute)s"),
                    ("2nd-non-existing-attribute", "non_existing.dotted.attribute"),
                    ("3rd-non-existing-attribute", 'some_var'),
                    (
                        "3rd-non-existing-attribute",
                        'Some random text, 1.2 float. Should not be added.'
                    ),
                ]),
                remove_empty=True,
                ignore_missing=True
            )
            logger.info('Composed message %s', 'remove non existing attribute')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([("level", "INFO"),
                        ("message", "Composed message remove non existing attribute")])
        )

    def test_ignore_trailing_dot_in_key(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ("level", "levelname"),
                    ("message", "message"),
                    ("trailing-dotted-key", "dotted_key."),
                ]),
                remove_empty=False,
                ignore_missing=True
            )
            logger.info('Composed message %s', 'remove non existing dotted key')
            logger.info(
                'Composed message %s',
                'existing dotted key as dict',
                extra={"dotted_key": {
                    "a": 12, "b": "this is b"
                }}
            )
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([("level", "INFO"),
                        ("message", "Composed message remove non existing dotted key"),
                        ("trailing-dotted-key", dictionary())])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([("level", "INFO"),
                        ("message", "Composed message existing dotted key as dict"),
                        ("trailing-dotted-key", dictionary([("a", 12), ("b", "this is b")]))])
        )

    def test_ignore_double_trailing_dot_in_key(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ("level", "levelname"),
                    ("message", "message"),
                    ("trailing-dotted-key", "dotted_key.."),
                ]),
                remove_empty=False,
                ignore_missing=True
            )
            logger.info('Composed message %s', 'remove non existing dotted key')
            logger.info(
                'Composed message %s',
                'existing dotted key as dict',
                extra={"dotted_key": ['a', 'b']}
            )
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([("level", "INFO"),
                        ("message", "Composed message remove non existing dotted key"),
                        ("trailing-dotted-key", [])])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([("level", "INFO"),
                        ("message", "Composed message existing dotted key as dict"),
                        ("trailing-dotted-key", ['a', 'b'])])
        )

    def test_leave_empty(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger_with_const_attribute(logger, remove_empty=False)
            logger.info('Composed message %s', 'leave empty')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("app", "test_application"),
                ('empty_attr', ''),
                ("message", "Composed message leave empty"),
            ])
        )

    def test_non_serializable_extra(self):

        class TestObject:

            def __str__(self):
                return 'Test Object'

        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(logger, add_always_extra=True)
            logger.info(
                'Composed message %s',
                'with non serializable extra',
                extra=dictionary([('test-object', TestObject())])
            )
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("levelname", "INFO"),
                ('name', 'test_formatter'),
                ("message", "Composed message with non serializable extra"),
                ('test-object', 'Test Object'),
            ])
        )

    def test_sub_key(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    ('request', dictionary([('path', 'request.path')])),
                    ('message', 'message'),
                ]),
                ignore_missing=True
            )
            logger.info(
                'Composed message %s',
                'with extra request',
                extra={'request': {
                    'path': '/my/path', 'method': 'GET', 'scheme': 'https'
                }}
            )
            logger.info('Composed message %s', 'without extra request')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", dictionary([("path", "/my/path")])),
                ("message", "Composed message with extra request"),
            ])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", dictionary([("path", "")])),
                ("message", "Composed message without extra request"),
            ])
        )

    def test_sub_key_list(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    ('request', ['request.path', 'request.method']),
                    ('message', 'message'),
                ]),
                ignore_missing=True
            )
            logger.info(
                'Composed message %s',
                'with extra request',
                extra={'request': {
                    'path': '/my/path', 'method': 'GET', 'scheme': 'https'
                }}
            )
            logger.info('Composed message %s', 'without extra request')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", ["/my/path", "GET"]),
                ("message", "Composed message with extra request"),
            ])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", ['', '']),
                ("message", "Composed message without extra request"),
            ])
        )

    def test_sub_key_remove_empty(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    ('request', dictionary([('path', 'request.path')])),
                    ('message', 'message'),
                ]),
                remove_empty=True,
                ignore_missing=True
            )
            logger.info(
                'Composed message %s',
                'with extra request',
                extra={'request': {
                    'path': '/my/path', 'method': 'GET'
                }}
            )
            logger.info('Composed message %s', 'without extra request')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", dictionary([("path", "/my/path")])),
                ("message", "Composed message with extra request"),
            ])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Composed message without extra request"),
            ])
        )

    def test_sub_key_list_remove_empty(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    ('request', ['request.path', 'request.method']),
                    ('message', 'message'),
                ]),
                remove_empty=True,
                ignore_missing=True
            )
            logger.info(
                'Composed message %s',
                'with extra request',
                extra={'request': {
                    'path': '/my/path', 'method': 'GET', 'scheme': 'https'
                }}
            )
            logger.info('Composed message %s', 'without extra request')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("request", ["/my/path", "GET"]),
                ("message", "Composed message with extra request"),
            ])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Composed message without extra request"),
            ])
        )

    def test_missing_attribute(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    ('1-missing-attr', 'this is a constant'),
                    ('2-missing-attr', 'my_constant'),
                    ('3-missing-attr', 'This is a constant with dot.'),
                    ('4-missing-attr', 'my.constant%()s'),
                    ('5-missing-attr', 'my.constant.%()s'),
                    ('6-missing-attr', '%(This is a constant with dot.)s'),
                    ('7-missing-attr', '%(test.a.)s'),
                    ('my-extra', 'my_extra'),
                    ('message', 'message'),
                ]),
                remove_empty=False,
                ignore_missing=True
            )
            logger.info(
                'Composed message %s',
                'with extra and constants',
                extra={'my_extra': 'this is an extra'}
            )
            logger.info('Composed message %s', 'without extra')
        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ('1-missing-attr', ''),
                ('2-missing-attr', ''),
                ('3-missing-attr', {}),
                ('4-missing-attr', ''),
                ('5-missing-attr', ''),
                ('6-missing-attr', ''),
                ('7-missing-attr', ''),
                ("my-extra", "this is an extra"),
                ("message", "Composed message with extra and constants"),
            ])
        )
        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ('1-missing-attr', ''),
                ('2-missing-attr', ''),
                ('3-missing-attr', {}),
                ('4-missing-attr', ''),
                ('5-missing-attr', ''),
                ('6-missing-attr', ''),
                ('7-missing-attr', ''),
                ("my-extra", ""),
                ("message", "Composed message without extra"),
            ])
        )



from functools import partial
from logging import LogRecord
from logging import setLogRecordFactory

_dict_ignore_missing_types = {}


class _DictIgnoreMissing(dict):
    _dft_value = ''

    def __getitem__(self, __k):
        try:
            return super().__getitem__(__k)
        except KeyError:
            return self._dft_value


class LogRecordIgnoreMissing(LogRecord):
    '''LogRecord that don't raise ValueError exception when trying to access missing extra attribute

    Missing/unknown extra attribute will returns `''` when accessed via the `__dict__` attribute.

    This is particularly usefull when using a style formatter with attribute that might be missing
    from the log record; e.g. `"%(message)s - %(extra_attribute)s"`

    The default value for missing attribute can be changed via `__dft_value` parameter.
    '''

    def __init__(self, *args, **kwargs):
        __dft_value = kwargs.pop('__dft_value', '')
        __dft_value_hash = str(__dft_value)
        super().__init__(*args, **kwargs)
        if __dft_value_hash not in _dict_ignore_missing_types:
            _dict_ignore_missing_type_name = '_DictIgnoreMissing_{}'.format(
                len(_dict_ignore_missing_types)
            )
            _dict_ignore_missing_types[__dft_value_hash] = type(
                _dict_ignore_missing_type_name, (_DictIgnoreMissing,), {"_dft_value": __dft_value}
            )
        self.__dict__ = _dict_ignore_missing_types[__dft_value_hash](self.__dict__)


def set_log_record_ignore_missing_factory(dft_value=''):
    '''Globally change the log record factory to the LogRecordIgnoreMissing factory

    This new log record won't raise any exception on unknown/missing attribute access, but will
    return the `dft_value` instead.
    '''
    setLogRecordFactory(partial(LogRecordIgnoreMissing, __dft_value=dft_value))


def reset_log_record_factory():
    '''Reset the log record factory to the original one LogRecord.
    '''
    setLogRecordFactory(LogRecord)  # pragma: no cover



import json
import logging
import sys
import unittest
from collections import OrderedDict

from flask import Flask

from logging_utilities.filters.flask_attribute import FlaskRequestAttribute
from logging_utilities.formatters.json_formatter import JsonFormatter

# From python3.7, dict is ordered
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict

app = Flask(__name__)


class FlaskJsonFormatterTest(unittest.TestCase):

    @classmethod
    def _configure_logger(
        cls,
        logger,
        fmt=None,
        add_always_extra=False,
        remove_empty=False,
        ignore_missing=False,
        flask_attributes=None,
        style='%'
    ):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            formatter = JsonFormatter(
                fmt,
                add_always_extra=add_always_extra,
                remove_empty=remove_empty,
                ignore_missing=ignore_missing,
                style=style
            )
            handler.setFormatter(formatter)
            flask_attribute = FlaskRequestAttribute(attributes=flask_attributes)
            handler.addFilter(flask_attribute)

    def test_json_formatter_flask_no_context(self):

        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    (
                        'request',
                        dictionary([
                            ('path', '%(flask_request_path)s'),
                            ('headers', dictionary([('Accept', 'flask_request_headers.Accept')]))
                        ])
                    ),
                    ('message', 'message'),
                ]),
                remove_empty=True,
                ignore_missing=True,
                flask_attributes=['path', 'headers']
            )
            logger.info('Simple message')

        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Simple message"),
            ])
        )

    def test_json_formatter_flask(self):

        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    (
                        'request',
                        dictionary([
                            ('path', '%(flask_request_path)s'),
                            ('headers', dictionary([('Accept', 'flask_request_headers.Accept')]))
                        ]),
                    ),
                    ('message', 'message'),
                ]),
                remove_empty=False,
                ignore_missing=True,
                flask_attributes=['path', 'headers']
            )

            with app.test_request_context('/make_report/2017'):
                logger.info('Simple message')

            with app.test_request_context('/make_report/2017', headers={'Accept': '*/*'}):
                logger.info('Simple message with headers')

        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Simple message"),
                (
                    "request",
                    dictionary([("path", "/make_report/2017"),
                                ("headers", dictionary([("Accept", "")]))])
                ),
            ])
        )

        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Simple message with headers"),
                (
                    "request",
                    dictionary([("path", "/make_report/2017"),
                                ("headers", dictionary([("Accept", "*/*")]))])
                ),
            ])
        )

    def test_json_formatter_flask_remove_empty(self):

        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_logger(
                logger,
                fmt=dictionary([
                    ('level', 'levelname'),
                    (
                        'request',
                        dictionary([
                            ('path', '%(flask_request_path)s'),
                            ('headers', dictionary([('Accept', 'flask_request_headers.Accept')]))
                        ]),
                    ),
                    ('message', 'message'),
                ]),
                remove_empty=True,
                ignore_missing=True,
                flask_attributes=['path', 'headers']
            )

            with app.test_request_context('/make_report/2017'):
                logger.info('Simple message')

            with app.test_request_context('/make_report/2017', headers={'Accept': '*/*'}):
                logger.info('Simple message with headers')

        self.assertDictEqual(
            json.loads(ctx.output[0], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Simple message"),
                ("request", dictionary([("path", "/make_report/2017")])),
            ])
        )

        self.assertDictEqual(
            json.loads(ctx.output[1], object_pairs_hook=dictionary),
            dictionary([
                ("level", "INFO"),
                ("message", "Simple message with headers"),
                (
                    "request",
                    dictionary([("path", "/make_report/2017"),
                                ("headers", dictionary([("Accept", "*/*")]))])
                ),
            ])
        )



import logging
import unittest
from logging import Formatter
from urllib.parse import quote

from flask import Flask

from logging_utilities.filters.flask_attribute import FlaskRequestAttribute
from logging_utilities.log_record import reset_log_record_factory
from logging_utilities.log_record import set_log_record_ignore_missing_factory

app = Flask(__name__)

FLASK_DEFAULT_FMT = "%(levelname)s:%(message)s:%(flask_request_url)s:%(flask_request_json)s:" \
                    "%(flask_request_query_string)s"
FLASK_DEFAULT_ATTRIBUTES = ['url', 'method', 'headers', 'json', 'query_string']


class FlaskAttributeTest(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    @classmethod
    def _configure_flask_attribute(cls, logger, fmt, flask_attributes):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            flask_attribute = FlaskRequestAttribute(attributes=flask_attributes)
            handler.addFilter(flask_attribute)
            handler.setFormatter(Formatter(fmt))

    def test_empty_flask_attribute_no_context(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, FLASK_DEFAULT_FMT, FLASK_DEFAULT_ATTRIBUTES)
            reset_log_record_factory()
            with self.assertRaises((ValueError, KeyError)):
                logger.info('Simple message')

    def test_empty_flask_attribute_no_context_ignore(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, FLASK_DEFAULT_FMT, FLASK_DEFAULT_ATTRIBUTES)
            set_log_record_ignore_missing_factory()
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
            reset_log_record_factory()
        self.assertEqual(
            ctx.output,
            [
                'INFO:Simple message:::',
                'INFO:Composed message: this is a composed message:::',
                'INFO:Composed message with extra:::'
            ]
        )

    def test_flask_attribute_json_data(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, FLASK_DEFAULT_FMT, FLASK_DEFAULT_ATTRIBUTES)

            with app.test_request_context('/make_report/2017', data={'format': 'short'}):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

            with app.test_request_context('/make_report/2017', data=''):
                logger.info('Simple message')

            with app.test_request_context(
                '/make_report/2017', data='non json data', content_type='application/json'
            ):
                logger.info('Simple message')

            with app.test_request_context(
                '/make_report/2017', data='{}', content_type='application/json'
            ):
                logger.info('Simple message')

            with app.test_request_context(
                '/make_report/2017',
                data='{"jsonData": "this is a json data"}',
                content_type='application/json'
            ):
                logger.info('Simple message')
        self.assertEqual(
            ctx.output,
            [
                # pylint: disable=line-too-long
                'INFO:Simple message:http://localhost/make_report/2017::',
                'INFO:Composed message: this is a composed message:http://localhost/make_report/2017::',
                'INFO:Composed message with extra:http://localhost/make_report/2017::',
                'INFO:Simple message:http://localhost/make_report/2017::',
                "INFO:Simple message:http://localhost/make_report/2017:non json data:",
                'INFO:Simple message:http://localhost/make_report/2017:{}:',
                "INFO:Simple message:http://localhost/make_report/2017:{'jsonData': 'this is a json data'}:",
            ]
        )

    def test_flask_attribute_query_string(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, FLASK_DEFAULT_FMT, FLASK_DEFAULT_ATTRIBUTES)
            with app.test_request_context('/make_report/2017?param1=value1'):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

            with app.test_request_context('/make_report/2017?param1=value1&param2=value2'):
                logger.info('Simple message')

            with app.test_request_context(
                '/make_report/2017?param1={}'.format(quote("This a string ?"))
            ):
                logger.info('Simple message')

        self.assertEqual(
            ctx.output,
            [
                # pylint: disable=line-too-long
                'INFO:Simple message:http://localhost/make_report/2017?param1=value1::param1=value1',
                'INFO:Composed message: this is a composed message:http://localhost/make_report/2017?param1=value1::param1=value1',
                'INFO:Composed message with extra:http://localhost/make_report/2017?param1=value1::param1=value1',
                'INFO:Simple message:http://localhost/make_report/2017?param1=value1&param2=value2::param1=value1&param2=value2',
                'INFO:Simple message:http://localhost/make_report/2017?param1=This%20a%20string%20?::param1=This%20a%20string%20%3F',
            ]
        )

    def test_flask_attribute_data(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, "%(message)s:%(flask_request_data)s", ['data'])
            data = "arbitrary data <Could be xml=true/>"
            with app.test_request_context(
                '/make_report/2017', data=data, headers={"Content-Type": "text/plain"}
            ):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

        self.assertEqual(
            ctx.output,
            [
                'Simple message:{}'.format(data),
                'Composed message: this is a composed message:{}'.format(data),
                'Composed message with extra:{}'.format(data),
            ]
        )

    def test_flask_attribute_form(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(
                logger,
                "%(message)s:%(flask_request_form)s:%(flask_request_mimetype)s",
                ['form', 'mimetype']
            )
            form = {'key1': 'value2', 'key2': '234'}
            with app.test_request_context('/make_report/2017', data=form):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

        self.assertEqual(
            ctx.output,
            [
                'Simple message:{}:application/x-www-form-urlencoded'.format(form),
                'Composed message: this is a composed message:{}:application/x-www-form-urlencoded'.
                format(form),
                'Composed message with extra:{}:application/x-www-form-urlencoded'.format(form),
            ]
        )

    def test_flask_attribute_args(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(logger, "%(message)s:%(flask_request_args)s", ['args'])

            with app.test_request_context('/make_report/2017?key1=value1&key2={"a":2}'):
                logger.info('Simple message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

        self.assertEqual(
            ctx.output,
            [
                "Simple message:{'key1': 'value1', 'key2': '{\"a\":2}'}",
                "Composed message with extra:{'key1': 'value1', 'key2': '{\"a\":2}'}",
            ]
        )

    def test_flask_attribute_mimetype(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(
                logger,
                "%(message)s:%(flask_request_mimetype)s:%(flask_request_mimetype_params)s",
                ['mimetype', 'mimetype_params']
            )
            data = "arbitrary data <Could be xml=true/>"
            with app.test_request_context(
                '/make_report/2017', data=data, headers={"Content-Type": "text/plain"}
            ):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

            with app.test_request_context(
                '/make_report/2017',
                data=data,
                headers={"Content-Type": "text/plain; charset=utf-8"}
            ):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

        self.assertEqual(
            ctx.output,
            [
                'Simple message:text/plain:{}',
                'Composed message: this is a composed message:text/plain:{}',
                'Composed message with extra:text/plain:{}',
                "Simple message:text/plain:{'charset': 'utf-8'}",
                "Composed message: this is a composed message:text/plain:{'charset': 'utf-8'}",
                "Composed message with extra:text/plain:{'charset': 'utf-8'}",
            ]
        )

    def test_flask_attribute_view_args(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_flask_attribute(
                logger, "%(message)s:%(flask_request_view_args)s", ['view_args']
            )

            # Handler required by add_url_rule function
            def handle_time(time):
                return

            # Request without view_args
            with app.test_request_context('/make_report'):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

            # Request with view args
            app.add_url_rule('/make_report/<time>', view_func=handle_time)
            with app.test_request_context('/make_report/current'):
                logger.info('Simple message')
                logger.info('Composed message: %s', 'this is a composed message')
                logger.info('Composed message %s', 'with extra', extra={'extra1': 23})

        self.assertEqual(
            ctx.output,
            [
                'Simple message:{}',
                'Composed message: this is a composed message:{}',
                'Composed message with extra:{}',
                "Simple message:{'time': 'current'}",
                "Composed message: this is a composed message:{'time': 'current'}",
                "Composed message with extra:{'time': 'current'}",
            ]
        )



import logging
from logging import LogRecord
from typing import List

from logging_utilities.thread_context import thread_context


class AddThreadContextFilter(logging.Filter):
    """Add local thread attributes to the log record.
    """

    def __init__(self, contexts: List[dict] = None) -> None:
        """Initialize the filter

        Args:
            contexts (List[dict], optional):
                List of values to add to the log record. Dictionary must contain value for
                'context_key' to read value from thread local variable. Dictionary must also contain
                'logger_key' to set the value on the log record.
        """
        self.contexts: List[dict] = [] if contexts is None else contexts
        super().__init__()

    def filter(self, record: LogRecord) -> bool:
        for ctx in self.contexts:
            if getattr(thread_context, ctx['context_key'], None) is not None:
                setattr(record, ctx['logger_key'], getattr(thread_context, ctx['context_key']))
        return True



import unittest

from django.conf import settings
from django.test import RequestFactory

from logging_utilities.django_middlewares.add_request_context import \
    AddToThreadContextMiddleware
from logging_utilities.thread_context import thread_context

if not settings.configured:
    settings.configure()


class AddToThreadContextMiddlewareTest(unittest.TestCase):

    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_add_request(self):

        def test_handler(request):
            r_from_var = getattr(thread_context, 'request', None)
            self.assertEqual(request, r_from_var)

        request = self.factory.get("/some_path?test=some_value")
        middleware = AddToThreadContextMiddleware(test_handler)
        middleware(request)



import json
import logging
import logging.config
import re
import sys
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from collections.abc import MutableMapping
from functools import partial
from logging import BASIC_FORMAT
from logging import PercentStyle as _PercentStyle
from logging import StrFormatStyle as _StrFormatStyle
from logging import StringTemplateStyle as _StringTemplateStyle

from logging_utilities.formatters import RECORD_DFT_ATTR
from logging_utilities.log_record import _DictIgnoreMissing
from logging_utilities.log_record import set_log_record_ignore_missing_factory

if sys.version_info.major < 3:
    raise ImportError('Only python 3 is supported')  # pragma: no cover

# From python3.7, dict is ordered. Ordered dict are preferred in order to keep the json output
# in the same order as its definition
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict  # pragma: no cover

DEFAULT_FORMAT = dictionary([('levelname', 'levelname'), ('name', 'name'), ('message', 'message')])


def _flatten_dict_gen(dct, parent_key, sep):
    for key, value in dct.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, MutableMapping):
            yield from flatten_dict(value, new_key, sep=sep).items()
        else:
            yield new_key, value


def flatten_dict(dct, parent_key='', sep='.'):
    return dict(_flatten_dict_gen(dct, parent_key, sep))


def is_style_format_valid(style):
    try:
        style.validate()
        return True
    except ValueError:
        return False


if sys.version_info.major >= 3 and sys.version_info.minor > 7:
    StrFormatStyle = _StrFormatStyle
    StringTemplateStyle = _StringTemplateStyle
    PercentStyle = _PercentStyle
else:
    # pragma: no cover
    # Python version prior 3.8 don't have a validate() method on the Style class.
    class ValidateStyleMixin():

        def validate(self):
            if not self.validation_pattern.search(self._fmt):
                raise ValueError(
                    "Invalid format '%s' for '%s' style" % (self._fmt, self.default_format[0])
                )

    class PercentStyle(_PercentStyle, ValidateStyleMixin):
        validation_pattern = re.compile(
            r'%\(\w+\)[#0+ -]*(\*|\d+)?(\.(\*|\d+))?[diouxefgcrsa%]', re.I
        )

    class StrFormatStyle(_StrFormatStyle, ValidateStyleMixin):
        validation_pattern = re.compile(
            r'{[a-z_]\w*(![rsa])?(:[#0+ -]*(\*|\d+)?(\.(\*|\d+))?[diouxefgcrsa%])?}', re.I
        )

    class StringTemplateStyle(_StringTemplateStyle, ValidateStyleMixin):
        validation_pattern = re.compile(r'{[a-z_]\w*}', re.I)


class EnhancedPercentStyle(PercentStyle):
    validation_pattern = re.compile(
        r'%\([\w\.]+\)[#0+ -]*(\*|\d+)?(\.(\*|\d+))?[diouxefgcrsa%]', re.I
    )

    def __init__(self, fmt, ignore_missing=False):
        super().__init__(fmt)
        self.ignore_missing = ignore_missing

    def _format(self, record):
        dct = {**record.__dict__, **flatten_dict(record.__dict__)}
        if self.ignore_missing:
            return self._fmt % _DictIgnoreMissing(dct)
        return self._fmt % dct


_ENHANCED_STYLES = {
    '%': (EnhancedPercentStyle, BASIC_FORMAT),
    '{': (StrFormatStyle, '{levelname}:{name}:{message}'),
    '$': (StringTemplateStyle, '${levelname}:${name}:${message}'),
}


class JsonFormatter(logging.Formatter):
    """Logging JSON Formatter

    This formatter transform the log message into a JSON object.
    """

    def __init__(
        self,
        fmt=None,
        datefmt=None,
        style='%',
        add_always_extra=False,
        filter_attributes=None,
        remove_empty=False,
        ignore_missing=False,
        **kwargs
    ):
        """JSON Formatter constructor

        Args:
            fmt: (dict)
                Format for the json output. It can be a string representing a json object or
                a dictionary. The object defines the JSON output:
                    KEY   := key to use in the json output
                    VALUE := value to use in the json output, this can be either a static string,
                             a record attribute name or a format string
            datefmt: (string)
                Format string for the date format
            style: (string)
                Use a style parameter of '%', '{' or '$' to specify that you want to use one of
                %-formatting, :meth:`str.format` (``{}``) formatting or :class:`string.Template`
                formatting in your format string.
            add_always_extra: (bool)
                If true the extra dictionary passed to the log function will be always
                added to the json output. Otherwise it is only added if the keys are in the
                fmt.
            filter_attributes: (list)
                List of record attributes added by a filter or log addapter.
                This list is used to differentiate record attributes from log extra
                attributes.
            remove_empty: (bool)
                If true the empty list, object or string in the output message are removed
            ignore_missing: (bool)
                If True, then all extra attributes from the log record that are missing (accessed
                by the fmt parameter) will be replaced by an empty string instead of raising a
                ValueError exception.
                NOTE: This has an impact on all formater not only on this one,
                see log_record.LogRecordIgnoreMissing.
            kwargs:
                Additional parameters passed to json.dumps().

        Raises:
            TypeError:  When the fmt parameter is in a wrong type
            json.decoder.JSONDecodeError: When fmt is a string that don't describe a json object
        """
        super().__init__(datefmt=datefmt, style=style)

        if fmt is None:
            fmt = DEFAULT_FORMAT
        if style == '%':
            self._style_constructor = partial(
                _ENHANCED_STYLES[style][0], ignore_missing=ignore_missing
            )
        else:
            self._style_constructor = _ENHANCED_STYLES[style][0]
        self._use_time = str(fmt).find('asctime') >= 0
        self.json_fmt = self._parse_fmt(fmt)
        self.add_always_extra = add_always_extra
        self.filter_attributes = filter_attributes
        self.remove_empty = remove_empty

        # support for `json.dumps` parameters
        self.kwargs = kwargs

        self.ignore_missing = ignore_missing
        if ignore_missing:
            set_log_record_ignore_missing_factory()

    @classmethod
    def _parse_fmt(cls, fmt):
        if isinstance(fmt, str):
            return json.loads(fmt, object_pairs_hook=dictionary)
        if isinstance(fmt, (dictionary, OrderedDict)):
            return fmt
        if isinstance(fmt, dict):  # pragma no cover
            warnings.warn(
                "Current python version is lower than 3.7.0, the key's order of dict may be "
                "different than the definition, please use `OrderedDict` instead.",
                UserWarning
            )
            return dictionary((k, fmt[k]) for k in sorted(fmt.keys()))

        raise TypeError(
            '`{}` type is not supported, `fmt` must be json `str`, `OrderedDict` or `dict` type.'.
            format(type(fmt))
        )  # pragma: no cover

    @classmethod
    def _add_extra_to_message(cls, extra, message):
        for key, value in extra.items():
            message[key] = value

    def _get_extra_attrs(self, record):

        def is_extra_attribute(attr):
            if attr not in RECORD_DFT_ATTR and (
                self.filter_attributes is None or attr not in self.filter_attributes
            ):
                return True
            return False

        extras = {key: record.__dict__[key] for key in record.__dict__ if is_extra_attribute(key)}
        if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
            return extras
        return dictionary((key, extras[key]) for key in sorted(extras.keys()))  # pragma no cover

    def _add_list_to_message(self, record, lst, message):
        for value in lst:
            if isinstance(value, (dict, OrderedDict)):
                intermediate_msg = dictionary()
                self._add_object_to_message(record, value, intermediate_msg)
                if not self.remove_empty or len(intermediate_msg) > 0:
                    message.append(intermediate_msg)
            elif isinstance(value, list):
                intermediate_msg = list()
                self._add_list_to_message(record, value, intermediate_msg)
                if not self.remove_empty or len(intermediate_msg) > 0:
                    message.append(intermediate_msg)
            elif value == 'exc_info':
                # exc_info might be true or contain exception info. We add it here only a boolean to
                # know if the message has exception info or not. The info itself will be in
                # 'exc_text' and is always appended to the json output when not available
                message.append(bool(record.exc_info))
            elif value in record.__dict__:
                intermediate_msg = getattr(record, value, None)
                if not self.remove_empty or intermediate_msg is not None:
                    message.append(intermediate_msg)
            elif isinstance(value, str):
                intermediate_msg = self._get_string_key_value(record, value)
                if self.remove_empty and not intermediate_msg:
                    pass
                else:
                    message.append(intermediate_msg)
            else:
                raise ValueError(
                    'Invalid value type={} for value={} in fmt'.format(type(value), value)
                )  # pragma no cover

    def _add_object_to_message(self, record, obj, message):
        for key, value in obj.items():
            if isinstance(value, (dict, OrderedDict)):
                message[key] = dictionary()
                self._add_object_to_message(record, value, message[key])
                if self.remove_empty and len(message[key]) == 0:
                    del message[key]
            elif isinstance(value, list):
                message[key] = list()
                self._add_list_to_message(record, value, message[key])
                if self.remove_empty and len(message[key]) == 0:
                    del message[key]
            elif value == 'exc_info':
                # exc_info might be true or contain exception info. We add it here only a boolean to
                # know if the message has exception info or not. The info itself will be in
                # 'exc_text' and is always appended to the json output when not available
                message[key] = bool(record.exc_info)
            elif value in record.__dict__:
                message[key] = getattr(record, value, '')
                if self.remove_empty and message[key] == '':
                    del message[key]
            elif isinstance(value, str):
                message[key] = self._get_string_key_value(record, value)
                if self.remove_empty and not message[key]:
                    del message[key]
            else:
                raise ValueError(
                    'Invalid value type={} for value={} in fmt'.format(type(value), value)
                )  # pragma no cover

    def _get_dotted_key_value(self, record, str_value):
        default_value = ''
        if str_value.endswith('..'):
            # if the dotted key has two trailing dot, this means that the value must be a list
            # therefore set the default value to an empty list
            default_value = []
        elif str_value.endswith('.'):
            # if the dotted key has a trailing dot, this mean that the value must be a dict
            # therefore set the default value to an empty dictionary.
            default_value = dictionary()

        def get_dotted_key(dct, dotted_key):
            if not isinstance(dct, (Mapping)):
                raise ValueError(
                    'Cannot get dotted key "{}" from "{}": '.format(dotted_key, dct) +
                    'is not a record or dictionary'
                )  # pragma: no cover
            key = dotted_key
            next_dotted_key = None
            if '.' in dotted_key:
                key, next_dotted_key = dotted_key.split('.', maxsplit=1)
                if next_dotted_key not in ['', '.']:
                    return get_dotted_key(dct.get(key, dictionary()), next_dotted_key)
            if self.ignore_missing:
                return dct.get(key, default_value)
            try:
                return dct[key]
            except KeyError as error:
                raise ValueError('Key "{}" not found in log record'.format(key)) from error

        return get_dotted_key(record.__dict__, str_value)

    def _get_string_key_value(self, record, value):
        style = self._style_constructor(value)

        if is_style_format_valid(style):
            # The value contains a valid style formatting (e.g. %(asctime)s)
            # therefore use the style formatter.
            return style.format(record)

        # Otherwise try to get a dotted key from the record
        return self._get_dotted_key_value(record, value)

    def usesTime(self):
        """
        Check if the format uses the creation time of the record.
        """
        return self._use_time

    def formatMessage(self, record):
        message = dictionary()
        self._add_object_to_message(record, self.json_fmt, message)
        return message

    def format(self, record):
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        if self.add_always_extra:
            extra = self._get_extra_attrs(record)

        message = self.formatMessage(record)

        if self.add_always_extra:
            self._add_extra_to_message(extra, message)

        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            message['exc_text'] = record.exc_text

        if record.stack_info:
            message['stack_info'] = self.formatStack(record.stack_info)

        # When adding all extras, to avoid crash when a log message adds an extra with a non
        # serializable object, we add a default serializer.
        default = self.kwargs.pop('default', None)
        if self.add_always_extra and default is None:
            default = str

        return json.dumps(message, default=default, **self.kwargs)


def basic_config(**kwargs):  # pragma: no cover
    """
    Do basic configuration for the logging system.

    This function does nothing if the root logger already has handlers
    configured. It is a convenience method intended for use by simple scripts
    to do one-shot configuration of the logging package.

    The default behavior is to create a StreamHandler which writes to
    sys.stderr, set a formatter using the BASIC_FORMAT format string, and
    add the handler to the root logger.

    A number of optional keyword arguments may be specified, which can alter
    the default behavior.

    Args:
        filename:
            Specifies that a FileHandler be created, using the specified filename,
            rather than a StreamHandler.
        filemode:
            Specifies the mode to open the file, if filename is specified
            (if filemode is unspecified, it defaults to 'a').
        format:
            Use the specified format string for the handler.
        datefmt:
            Use the specified date/time format.
        style:
            If a format string is specified, use this to specify the
            type of format string (possible values '%', '{', '$', for
            %-formatting, :meth:`str.format` and :class:`string.Template`
            - defaults to '%').
        level:
            Set the root logger level to the specified level.
        stream:
            Use the specified stream to initialize the StreamHandler. Note
            that this argument is incompatible with 'filename' - if both
            are present, 'stream' is ignored.
        handlers:
            If specified, this should be an iterable of already created
            handlers, which will be added to the root handler. Any handler
            in the list which does not have a formatter assigned will be
            assigned the formatter created in this function.

    Note that you could specify a stream created using open(filename, mode)
    rather than passing the filename and mode in. However, it should be
    remembered that StreamHandler does not close its stream (since it may be
    using sys.stdout or sys.stderr), whereas FileHandler closes its stream
    when the handler is closed.
    """
    # Add thread safety in case someone mistakenly calls
    # basic_config() from multiple threads
    # pylint: disable=protected-access
    logging._acquireLock()
    try:
        if len(logging.root.handlers) > 0:
            return

        handlers = kwargs.pop("handlers", None)
        if handlers is None:
            if "stream" in kwargs and "filename" in kwargs:
                raise ValueError("'stream' and 'filename' should not be specified together")
        else:
            if "stream" in kwargs or "filename" in kwargs:
                raise ValueError(
                    "'stream' or 'filename' should not be specified together with 'handlers'"
                )
        if handlers is None:
            filename = kwargs.pop("filename", None)
            mode = kwargs.pop("filemode", 'a')
            if filename:
                handler = logging.FileHandler(filename, mode)
            else:
                stream = kwargs.pop("stream", None)
                handler = logging.StreamHandler(stream)
            handlers = [handler]
        dfs = kwargs.pop("datefmt", None)
        style = kwargs.pop("style", '%')
        if style not in logging._STYLES:
            raise ValueError('Style must be one of: %s' % ','.join(logging._STYLES.keys()))
        fmt = kwargs.pop("format", DEFAULT_FORMAT)
        formatter = JsonFormatter(fmt, dfs, style)
        for handler in handlers:
            if handler.formatter is None:
                handler.setFormatter(formatter)
            logging.root.addHandler(handler)
        level = kwargs.pop("level", None)
        if level is not None:
            logging.root.setLevel(level)
        if kwargs:
            keys = ', '.join(kwargs.keys())
            raise ValueError('Unrecognized argument(s): %s' % keys)
    finally:
        logging._releaseLock()



import json
import logging
import unittest

from logging_utilities.filters.attr_type_filter import AttrTypeFilter
from logging_utilities.filters.attr_type_filter import is_instance
from logging_utilities.formatters.json_formatter import JsonFormatter


class TestObject:

    def __str__(self):
        return 'Test Object'


class TestSubobject(TestObject):

    def __str__(self):
        return 'Test Subobject'


class IsInstanceTest(unittest.TestCase):

    def test_pass_types(self):
        self.assertTrue(is_instance(TestObject(), TestObject))
        self.assertFalse(is_instance(TestObject(), TestSubobject))
        self.assertTrue(is_instance(TestSubobject(), TestObject))
        self.assertTrue(is_instance(TestSubobject(), TestSubobject))
        self.assertFalse(is_instance(TestSubobject(), str))

    def test_pass_dotless_strings(self):
        self.assertTrue(is_instance(TestObject(), 'TestObject'))
        self.assertFalse(is_instance(TestObject(), 'TestSubobject'))
        self.assertTrue(is_instance(TestSubobject(), 'TestObject'))
        self.assertTrue(is_instance(TestSubobject(), 'TestSubobject'))
        self.assertFalse(is_instance(TestSubobject(), 'str'))

    def test_pass_dotted_strings(self):
        self.assertTrue(is_instance(TestObject(), 'tests.test_attr_type_filter.TestObject'))
        self.assertFalse(is_instance(TestObject(), 'tests.test_attr_type_filter.TestSubobject'))
        self.assertTrue(is_instance(TestSubobject(), 'tests.test_attr_type_filter.TestObject'))
        self.assertTrue(is_instance(TestSubobject(), 'tests.test_attr_type_filter.TestSubobject'))
        self.assertFalse(is_instance(TestSubobject(), 'builtins.str'))


class AttrTypeFilterTest(unittest.TestCase):

    def setUp(self):
        self.logger = logging.getLogger('test_formatter')
        self.logger.setLevel(logging.DEBUG)

    def log_and_assert(self, type_validators, extra_dict, filtered_extra_dict):
        with self.assertLogs(self.logger, level=logging.INFO) as ctx:
            for handler in self.logger.handlers:
                handler.setFormatter(JsonFormatter(add_always_extra=True))
                for validator in type_validators:
                    handler.addFilter(validator)
            self.logger.info('Composed message %s', 'with extra', extra=extra_dict)
        message = json.loads(ctx.output[0], object_pairs_hook=dict)
        self.assertDictEqual(
            message,
            {
                "levelname": "INFO",
                "name": "test_formatter",
                "message": "Composed message with extra",
                **filtered_extra_dict
            },
        )

    def test_include_filter(self):
        self.log_and_assert(
            type_validators=[
                AttrTypeFilter({
                    'request': 'dict',
                    'unexistent': 'str',
                    'entry': bool,
                    'abc': ['bool', str, 'int', 'TestSubobject'],
                    'def': 'TestObject',
                    'ghi': 'TestObject',
                    'ghi2': 'tests.test_attr_type_filter.TestObject',
                    'jkl': ['bool', str, 'int', 'TestSubobject'],
                    'mno': 'TestSubobject'
                })
            ],
            extra_dict={
                'entry': 'hey',  #no match
                'abc': TestObject(),  #no match
                'def': TestObject(),  #match
                'ghi': TestSubobject(),  #match (checks that super class detection works)
                'ghi2': TestSubobject(),  #match
                'jkl': TestSubobject(),  #match
                'mno': TestObject(),  #no match
                'request': {  #match
                    'path': '/my/path', 'method': 'GET', 'comment': TestObject()
                }
            },
            filtered_extra_dict={
                "request": {
                    "path": '/my/path',
                    "method": "GET",
                    "comment": "Test Object"  #str() serializer is used if extra is not serializable
                },
                'def': 'Test Object',
                'ghi': 'Test Subobject',
                'ghi2': 'Test Subobject',
                'jkl': 'Test Subobject',
            }
        )

    def test_exclude_filter(self):
        self.log_and_assert(
            type_validators=[
                AttrTypeFilter(
                    is_blacklist=True,
                    typecheck_list={
                        'request': 'dict',
                        'unexistent': 'str',
                        'entry': bool,
                        'abc': ['bool', str, 'int'],
                    }
                )
            ],
            extra_dict={
                'entry': 'hey',  #no match
                'abc': TestObject(),  #no match
                'request': {  #match
                    'path': '/my/path', 'method': 'GET', 'comment': TestObject()
                }
            },
            filtered_extra_dict={
                'entry': 'hey', 'abc': 'Test Object'
            }
        )

    def test_include_and_exclude_filter(self):
        self.log_and_assert(
            type_validators=[
                AttrTypeFilter({
                    'entry': 'bool',
                    'entry3': 'int',
                    'request': 'builtins.dict',
                    'abc': 'tests.test_attr_type_filter.TestObject',
                    'unexistent': 'str',
                }),
                AttrTypeFilter(
                    is_blacklist=True,
                    typecheck_list={
                        'entry2': str,
                        'entry3': int,
                        'abc': [bool, str, int],
                        'nested': dict,
                    }
                )
            ],
            extra_dict={
                'entry': 'hey',  #no match, no filter => hidden
                'abc': TestObject(),  #no filter, no match => show
                "entry2": "hey2",  #no filter, match => hidden
                'request': {  #match, no filter => show
                    'path': '/my/path',
                    'scheme': 'https',
                },
                "entry3": "hey3",  #no match, no match => hidden
                'nested': {  #no filter, match => hidden
                    'param1': 'text1',
                    'param2': 'text2',
                }
            },
            filtered_extra_dict={
                'abc': 'Test Object', 'request': {
                    'path': '/my/path', 'scheme': 'https'
                }
            }
        )



import logging
import re
import sys
import warnings
from collections import OrderedDict

from django.http import HttpRequest

# From python3.7, dict is ordered. Ordered dict are preferred in order to keep the json output
# in the same order as its definition
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict


def _pattern(text):
    return r'^{}\b'.format(text.replace('.', '[.]'))


class JsonDjangoRequest(logging.Filter):
    """Convert Django request to a json object

    This filter recursively converts a django HttpRequest object to a python dictionary that can be
    dumped into a json string. This is useful for example if you want to log an extra parameter of
    type HttpRequest with the JSON formatter. If the specified attribute is not of type HttpRequest,
    it will simply be ignored and passed through.

    Additionally the attributes of the request that needs to be jsonify can be configured using the
    `include_keys` and/or `exclude_keys` parameters.

    The django framework adds sometimes an HttpRequest or socket object under "record.request" when
    logging. So if you decide to use the attribute name "request" for this filter, beware that you
    will need to handle the case where the attribute is of type 'socket' separately, for example by
    filtering it out using the attribute type filter. (see example in README)
    """

    def __init__(self, include_keys=None, exclude_keys=None, attr_name='http_request'):
        """Initialize the filter

        Args:
            include_keys: (list | None)
                All request attributes that match any of the dotted keys of the list will be added
                to the jsonifiable object. When None then all attributes are added
                (default behavior).
            exclude_keys: (list | None)
                All request attributes that match any of the dotted keys of the list will not be
                added to the jsonifiable object. NOTE this has precedence to include_keys
                which means that if a key is in both lists, then it is not added.
            attr_name: str
                The name of the attribute that stores the HttpRequest object. The default is
                'http_request'.
                (Note that django sometimes stores an "HttpRequest" under the attribute "request".
                This is however not the default as django also stores other types of objects under
                this attribute name.)
        """
        self.include_keys = include_keys
        self.exclude_keys = exclude_keys
        self.attr_name = attr_name
        super().__init__()

    def filter(self, record):
        if not hasattr(record, self.attr_name):
            return True

        self._jsonify_request(record)

        return True

    def _jsonify_request(self, record):
        orig_request = getattr(record, self.attr_name)
        if isinstance(orig_request, HttpRequest) and hasattr(orig_request, '__dict__'):
            request = self._jsonify_dict(self.attr_name, orig_request.__dict__)
            if self._add_key(self.attr_name + '.headers', 'headers'):
                # HttpRequest has a special headers property that is cached and is not always in
                # record.http_request.__dict__
                request['headers'] = self._jsonify_dict(
                    self.attr_name + '.headers', orig_request.headers
                )
            setattr(record, self.attr_name, request)

    def _jsonify_dict(self, prefix, dct):
        json_obj = dictionary()
        if sys.version_info.major < 3 and sys.version_info.minor < 7:
            dct = OrderedDict(sorted(dct.items(), key=lambda t: t[0]))
        for key, value in dct.items():
            dotted_key = '{}.{}'.format(prefix, key)
            if not self._add_key(dotted_key, key):
                continue
            if hasattr(value, '__dict__'):
                json_obj[key] = self._jsonify_dict(dotted_key, value.__dict__)
            elif isinstance(value, dict):
                json_obj[key] = self._jsonify_dict(dotted_key, value)
            elif isinstance(value, (str, int, float, type(None), bool, tuple, list)):
                json_obj[key] = value
            elif isinstance(value, (bytes)):
                json_obj[key] = str(value)
            else:
                warnings.warn(
                    "Cannot jsonify key {} with value {}: unsupported type={}".format(
                        dotted_key, value, type(value)
                    )
                )
                json_obj[key] = str(value)
        return json_obj

    def _add_key(self, dotted_key, key):
        return self._include_key(dotted_key, key) and not self._exclude_key(dotted_key, key)

    def _include_key(self, dotted_key, key):
        if self.include_keys is None:
            # if no include_keys is configured add all keys except for private
            if key.startswith('_'):
                return False
            return True

        def match(item):
            return (
                re.match(_pattern(dotted_key), item) or
                (re.match(_pattern(item), dotted_key) and not key.startswith('_'))
            )

        return any(map(match, self.include_keys))

    def _exclude_key(self, dotted_key, key):
        if self.exclude_keys is None:
            # if no exclude_keys is configured only exclude private key
            if key.startswith('_'):
                return True
            return False

        def match(item):
            return re.match(_pattern(item), dotted_key)

        return any(map(match, self.exclude_keys))



from logging_utilities.thread_context import thread_context


class AddToThreadContextMiddleware(object):
    """Django middleware that stores request to thread local variable.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        setattr(thread_context, 'request', request)
        response = self.get_response(request)
        setattr(thread_context, 'request', None)
        return response



import json
import logging
import sys
import unittest
from collections import OrderedDict

from django.conf import settings
from django.test import RequestFactory

from logging_utilities.filters.add_thread_context_filter import \
    AddThreadContextFilter
from logging_utilities.formatters.json_formatter import JsonFormatter
from logging_utilities.thread_context import thread_context

if not settings.configured:
    settings.configure()

# From python3.7, dict is ordered
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict

logger = logging.getLogger(__name__)


class AddThreadContextFilterTest(unittest.TestCase):

    def setUp(self) -> None:
        self.factory = RequestFactory()

    @classmethod
    def _configure_json_filter(cls, _logger):
        _logger.setLevel(logging.DEBUG)
        for handler in _logger.handlers:
            handler.setFormatter(JsonFormatter(add_always_extra=True))

    def test_add_thread_context_no_request(self):
        with self.assertLogs('test_logger', level=logging.DEBUG) as ctx:
            test_logger = logging.getLogger("test_logger")
            self._configure_json_filter(test_logger)
            test_logger.addFilter(
                AddThreadContextFilter(
                    contexts=[{
                        'logger_key': 'http_request', 'context_key': 'request'
                    }]
                )
            )
            test_logger.debug("some message")

        message1 = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message1,
            dictionary([("levelname", "DEBUG"), ("name", "test_logger"),
                        ("message", "some message")])
        )

    def test_add_thread_context(self):
        test_cases = [
            {
                'logger_name': 'test_1',
                'var_key': 'request',
                'var_val': "some value",
                'attr_name': 'http_request',
                'expect_value': "some value",
                'log_message': 'a log message has appeared',
            },
            {
                'logger_name': 'test_2',
                'var_key': 'request',
                'var_val': self.factory.get("/some_path"),
                'attr_name': 'request',
                'expect_value': "<WSGIRequest: GET '/some_path'>",
                'log_message': 'another log message has appeared',
            },
        ]

        for tc in test_cases:
            with self.assertLogs(tc['logger_name'], level=logging.DEBUG) as ctx:
                test_logger = logging.getLogger(tc['logger_name'])
                setattr(thread_context, tc['var_key'], tc['var_val'])
                self._configure_json_filter(test_logger)
                test_logger.addFilter(
                    AddThreadContextFilter(
                        contexts=[{
                            'logger_key': tc['attr_name'], 'context_key': tc['var_key']
                        }]
                    )
                )

                test_logger.debug(tc['log_message'])
                setattr(thread_context, tc['var_key'], None)

            message1 = json.loads(ctx.output[0], object_pairs_hook=dictionary)
            self.assertDictEqual(
                message1,
                dictionary([("levelname", "DEBUG"), ("name", tc['logger_name']),
                            ("message", tc['log_message']), (tc['attr_name'], tc['expect_value'])])
            )



import logging
import unittest
from logging import Formatter

from nose2.tools import params

from logging_utilities.filters import ConstAttribute


class RecordAttributesIgnoreMissingTest(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    @classmethod
    def _configure_const_attribute(cls, logger, fmt, value):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            const_attribute = ConstAttribute(const_key=value)
            handler.addFilter(const_attribute)
            handler.setFormatter(Formatter(fmt))

    @params('string value', 1, 1.1, [1, 'b'], None, {}, {'a': 1, 'b': {'c': ['a']}})
    def test_const_attribute(self, value):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_const_attribute(
                logger, "%(levelname)s:%(const_key)s:%(message)s", value
            )
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
        self.assertEqual(
            ctx.output,
            [
                'INFO:{}:Simple message'.format(value),
                'INFO:{}:Composed message: this is a composed message'.format(value),
                'INFO:{}:Composed message with extra'.format(value)
            ]
        )

    @params('string value', 1, 1.1, [1, 'b'], None, {}, {'a': 1, 'b': {'c': ['a']}})
    def test_const_attribute_and_extra(self, value):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            logger = logging.getLogger('test_formatter')
            self._configure_const_attribute(
                logger, "%(levelname)s:%(const_key)s:%(message)s:%(extra1)s", value
            )
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
        self.assertEqual(ctx.output, ['INFO:{}:Composed message with extra:23'.format(value)])



import json
import logging
import sys
import time
import unittest
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor

from django.conf import settings
from django.test import RequestFactory

from logging_utilities.django_middlewares.add_request_context import \
    AddToThreadContextMiddleware
from logging_utilities.filters.add_thread_context_filter import \
    AddThreadContextFilter
from logging_utilities.filters.django_request import JsonDjangoRequest
from logging_utilities.formatters.json_formatter import JsonFormatter

# From python3.7, dict is ordered
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
    dictionary = dict
else:
    dictionary = OrderedDict

if not settings.configured:
    settings.configure()

logger = logging.getLogger(__name__)


class AddRequestToLogTest(unittest.TestCase):

    def setUp(self) -> None:
        self.factory = RequestFactory()

    @classmethod
    def _configure_django_filter(cls, _logger):
        for handler in _logger.handlers:
            handler.setFormatter(JsonFormatter(add_always_extra=True))
            handler.addFilter(
                AddThreadContextFilter(
                    contexts=[{
                        'logger_key': 'request', 'context_key': 'request'
                    }]
                )
            )
            handler.addFilter(
                JsonDjangoRequest(
                    include_keys=["request.path", "request.META.QUERY_STRING"], attr_name="request"
                )
            )

    def test_log_request_context(self):

        def test_handler(request):
            logger.info("some value")
            return "some response"

        with self.assertLogs(logger, level=logging.DEBUG) as ctx:
            # Global config of filter
            self._configure_django_filter(logger)
            request = self.factory.get("/some_path?test=some_value")
            middleware = AddToThreadContextMiddleware(test_handler)
            middleware(request)

        print(ctx.output[0])
        message1 = json.loads(ctx.output[0], object_pairs_hook=dictionary)
        self.assertDictEqual(
            message1,
            dictionary([
                ("levelname", "INFO"),
                ("name", "tests.test_log_request_context"),
                ("message", "some value"),
                (
                    "request",
                    dictionary([("path", "/some_path"),
                                ("META", dictionary([("QUERY_STRING", "test=some_value")]))])
                ),
            ])
        )


class MultiprocessLoggingTest(unittest.TestCase):

    def setUp(self) -> None:
        self.factory = RequestFactory()

    @classmethod
    def _configure_django_filter(cls, _logger):
        for handler in _logger.handlers:
            handler.setFormatter(JsonFormatter(add_always_extra=True))
            handler.addFilter(
                AddThreadContextFilter(
                    contexts=[{
                        'logger_key': 'request', 'context_key': 'request'
                    }]
                )
            )
            handler.addFilter(JsonDjangoRequest(include_keys=["request.path"], attr_name="request"))

    def test_threaded_logging(self):

        def test_handler(request):
            time.sleep(1)
            logger.info(request.path)
            return "some response"

        paths = [
            "/first_path",
            "/second_path",
            "/third_path",
        ]

        def execute_request(path):
            request = self.factory.get(path)
            middleware = AddToThreadContextMiddleware(test_handler)
            middleware(request)

        with self.assertLogs(logger, level=logging.DEBUG) as ctx:
            # Global config of filter
            self._configure_django_filter(logger)
            with ThreadPoolExecutor() as executor:
                futures = []
                for path in paths:
                    futures.append(executor.submit(execute_request, path))

        for output in ctx.output:
            msg = json.loads(output, object_pairs_hook=dictionary)
            self.assertEqual(msg["message"], msg["request"]["path"])



import logging

from werkzeug.datastructures import ImmutableDict
from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.datastructures import MultiDict
from werkzeug.exceptions import HTTPException

from flask import has_request_context
from flask import request


class FlaskRequestAttribute(logging.Filter):
    """Logging Flask attributes record

    This filter adds Flask request context attributes to the log record.

    Flask request attributes are added as record attributes with the 'flask_request_' prefix.
    """

    def __init__(self, attributes=None):
        """Initialize the filter

        Args:
            attributes: (list)
                Flask request attribute names list to add to the log record
        """
        super().__init__()
        self.attributes = attributes if attributes else list()

    def filter(self, record):
        for attribute in self.attributes:
            if has_request_context():
                rec_attribute = 'flask_request_' + attribute
                try:
                    value = getattr(request, attribute)
                except HTTPException:
                    # accessing the request.json might raise an HTTPException if the request
                    # is malformed for json data. In this case we don't want the filter to crash
                    # but simply set an empty value
                    if attribute == 'json':
                        if isinstance(request.data, bytes):
                            value = request.data.decode('utf-8')
                        else:
                            value = str(request.data)
                    else:
                        raise
                # Accessing flask_request_view_args.<key> might rise an exception if
                # flask_request_view_args is Null. To safely access flask_request_view_args
                # None is replaced by an empty dict.
                if attribute == 'view_args' and value is None:
                    setattr(record, rec_attribute, {})
                elif isinstance(value, (ImmutableDict, ImmutableMultiDict, MultiDict)):
                    setattr(record, rec_attribute, dict(value))
                elif value is None or isinstance(value, (str, int, float, dict, list)):
                    setattr(record, rec_attribute, value)
                elif isinstance(value, bytes):
                    setattr(record, rec_attribute, value.decode('utf-8'))
                elif attribute == 'headers':
                    setattr(record, rec_attribute, dict(value.items()))
                else:
                    raise ValueError('Attribute %s=%s unsupported type' % (attribute, value))
        return True



import logging
from functools import wraps

from .thread_context import ThreadMappingContext

__record_factory_wrapped = None  # pylint: disable=invalid-name
__initial_record_factory = logging.getLogRecordFactory()
__context = None  # pylint: disable=invalid-name


def set_logging_context(context=None):
    '''Set a logging context

    The context is set per thread (each thread can have different context) and is set to every
    log record in the `context` attribute.

    Args:
        context: (dict, None)
            Context to set, by default `None`. The context can be later retrieved and modified using
            `get_logging_context()`
    '''
    global __record_factory_wrapped  # pylint: disable=global-statement, invalid-name
    global __context  # pylint: disable=global-statement, invalid-name
    current_factory = logging.getLogRecordFactory()
    if __context is None:
        __context = ThreadMappingContext()
    __context.init(context)
    if current_factory != __record_factory_wrapped:
        __initial_record_factory = logging.getLogRecordFactory()
        __record_factory_wrapped = __wrap_log_record_with_context(current_factory, __context)
        logging.setLogRecordFactory(__record_factory_wrapped)


def get_logging_context():
    '''Return the current logging context if set or `None` otherwise'''
    return __context


def remove_logging_context():
    '''Remove the logging context'''
    global __context  # pylint: disable=global-statement, invalid-name
    logging.setLogRecordFactory(__initial_record_factory)
    __context = None


def __wrap_log_record_with_context(record_factory, context):

    @wraps(record_factory)
    def wrapper(*args, **kwargs):
        record = record_factory(*args, **kwargs)
        record.context = context
        return record

    return wrapper



import logging


def is_instance(attr, of_type):
    if isinstance(of_type, str):
        attr_class = type(attr)
        if '.' in of_type:
            return (
                '{}.{}'.format(attr_class.__module__, attr_class.__name__) == of_type or
                of_type in ['{}.{}'.format(c.__module__, c.__name__) for c in attr_class.__bases__]
            )
        return attr_class.__name__ == of_type or of_type in [
            c.__name__ for c in attr_class.__bases__
        ]
    return isinstance(attr, of_type)


class AttrTypeFilter(logging.Filter):
    """Filter attributes based on their types

    This filter can help in case multiple libraries/frameworks etc. use the
    same extra properties in the extra parameter of the logging. It filters
    these extra properties by type, either with a whitelist (the default) or
    with a blacklist.
    """

    def __init__(self, typecheck_list, *, is_blacklist=False):
        """Initialize the filter

        Args:
            typecheck_list: dict(key, type|list of types)
                A dictionary that maps keys to a type or a list of types.
                By default, it will only keep a parameter matching a key
                if the types match or if any of the types in the list match
                (white list). If in black list mode, it will only keep a
                parameter if the types don't match. Parameters not appearing
                in the dict will be ignored and passed though regardless of the
                mode (whitelist or blacklist).
            is_blacklist: bool (default: false)
                Whether the list passed should be a blacklist or a whitelist.
                To use both, simply include this filter two times, one time with
                this parameter set true and one time with this parameter set false.
        """
        self.typecheck_list = typecheck_list
        for key in self.typecheck_list:
            if not isinstance(self.typecheck_list[key], list):
                self.typecheck_list[key] = [self.typecheck_list[key]]
        self.is_blacklist = is_blacklist
        super().__init__()

    def filter(self, record):
        for key, whitelisted_types in self.typecheck_list.items():
            if not hasattr(record, key):
                continue
            item = getattr(record, key)
            if any(is_instance(item, t) for t in whitelisted_types) is self.is_blacklist:
                delattr(record, key)

        return True



import unittest

from nose2.tools import params

from logging_utilities.formatters.json_formatter import EnhancedPercentStyle


class RegexTests(unittest.TestCase):

    @params(
        '%(asctime)s',
        '%(my.dotted.key)s',
        '%(my.dotted.key)s with constant',
        'leading constant %(my.dotted.key)s with constant',
        'constant%(my.dotted.key)sconstant',
        '%(my.dotted.key)d',
        '%(my.dotted.key)3.4d',
    )
    def test_json_formatter_enhanced_percent_style_regex_match(self, value):
        self.assertIsNotNone(EnhancedPercentStyle.validation_pattern.search(value))

    @params(
        'my.key',
        'my.key2',
        'my.',
        'k1.k3.',
        'k_1.',
        'k_.sk_2',
        '_my.key',
        '_1.',
        '_1._1',
        '_1a._2b._3b',
        'a.b.c.',
        'a1a.b2b.c2c.',
        'a1a_.b2_b.c_2c.',
        'my key.',
        '1.2',
        'This is a text. It contains dot.',
        'my.key ',
        ' my.key',
        'my.key-test',
        '1',
        '1.test',
        '_1.1test',
        '%()s',
        'asd %()s asdf'
    )
    def test_json_formatter_enhanced_percent_style_regex_not_match(self, value):
        self.assertIsNone(EnhancedPercentStyle.validation_pattern.search(value))



import logging
import unittest
from logging import Formatter
from logging import Logger

from nose2.tools import params

from logging_utilities.log_record import LogRecordIgnoreMissing
from logging_utilities.log_record import _DictIgnoreMissing
from logging_utilities.log_record import reset_log_record_factory
from logging_utilities.log_record import set_log_record_ignore_missing_factory


class DictIgnoreMissingTest(unittest.TestCase):

    def test_dict_ignore_missing_default(self):
        dct = _DictIgnoreMissing()
        self.assertEqual(dct['unknown'], '')

    @params(None, 'my-default', 0, 100, {'1': 'default'}, {}, [], [1], ['default'])
    def test_dict_ignore_missing(self, default):

        class _DictIgnoreMissingX(_DictIgnoreMissing):
            _dft_value = default

        dct = _DictIgnoreMissingX()
        self.assertEqual(dct['unknown'], default)


class LogRecordIgnoreMissingTest(unittest.TestCase):

    def test_log_record_missing_default(self):
        record = LogRecordIgnoreMissing(
            'my-record', logging.INFO, 'my/path', 0, 'my message', None, False
        )
        self.assertEqual(record.__dict__['unknown'], '')
        self.assertEqual(record.__dict__['name'], 'my-record')
        self.assertEqual(record.__dict__['levelno'], logging.INFO)

    def test_logger_missing_default(self):
        set_log_record_ignore_missing_factory()
        logger = Logger('my-logger')

        record = logger.makeRecord(
            'my-record',
            logging.INFO,
            'test',
            0,
            'my message: %s %d', ('test', 2),
            False,
            extra={'my-extra': 'this is an extra'}
        )

        self.assertEqual(record.__dict__['name'], 'my-record')
        self.assertEqual(record.__dict__['levelno'], logging.INFO)
        self.assertEqual(record.getMessage(), 'my message: test 2')
        self.assertEqual(record.__dict__['my-extra'], 'this is an extra')
        self.assertEqual(record.__dict__['unknown'], '')
        reset_log_record_factory()

    @params(None, 'my-default', 0, 100, {'1': 'default'}, {}, [], [1], ['default'])
    def test_logger_missing(self, default):
        set_log_record_ignore_missing_factory(dft_value=default)
        logger = Logger('my-logger')

        record = logger.makeRecord(
            'my-record',
            logging.INFO,
            'test',
            0,
            'my message: %s %d', ('test', 2),
            False,
            extra={'my-extra': 'this is an extra'}
        )

        self.assertEqual(record.__dict__['name'], 'my-record')
        self.assertEqual(record.__dict__['levelno'], logging.INFO)
        self.assertEqual(record.getMessage(), 'my message: test 2')
        self.assertEqual(record.__dict__['my-extra'], 'this is an extra')
        self.assertEqual(record.__dict__['unknown'], default)
        reset_log_record_factory()


class LoggerIgnoreMissingTest(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    @classmethod
    def _configure_logging(cls, logger, fmt):
        logger.setLevel(logging.DEBUG)

        for handler in logger.handlers:
            handler.setFormatter(Formatter(fmt))

    def test_logger_ignore_missing_extra_default(self):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            set_log_record_ignore_missing_factory()
            logger = logging.getLogger('test_formatter')
            self._configure_logging(logger, "%(levelname)s:%(message)s:%(missing_attribute)s")
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
            reset_log_record_factory()
        self.assertEqual(
            ctx.output,
            [
                'INFO:Simple message:',
                'INFO:Composed message: this is a composed message:',
                'INFO:Composed message with extra:'
            ]
        )

    @params(None, 'my-default', 0, 100, {'1': 'default'}, {}, [], [1], ['default'])
    def test_logger_ignore_missing_extra(self, default):
        with self.assertLogs('test_formatter', level=logging.DEBUG) as ctx:
            set_log_record_ignore_missing_factory(default)
            logger = logging.getLogger('test_formatter')
            self._configure_logging(logger, "%(levelname)s:%(message)s:%(missing_attribute)s")
            logger.info('Simple message')
            logger.info('Composed message: %s', 'this is a composed message')
            logger.info('Composed message %s', 'with extra', extra={'extra1': 23})
            reset_log_record_factory()
        self.assertEqual(
            ctx.output,
            [
                'INFO:Simple message:' + str(default),
                'INFO:Composed message: this is a composed message:' + str(default),
                'INFO:Composed message with extra:' + str(default)
            ]
        )



from abc import abstractmethod
from collections.abc import MutableMapping


class BaseContext(MutableMapping):
    __marker = object()

    @abstractmethod
    def init(self, data):
        pass  # pragma: no cover

    @abstractmethod
    def get(self, key, default=None):
        pass  # pragma: no cover

    @abstractmethod
    def pop(self, key, default=__marker):
        pass  # pragma: no cover

    @abstractmethod
    def set(self, key, value):
        pass  # pragma: no cover

    @abstractmethod
    def delete(self, key: str):
        pass  # pragma: no cover

    @abstractmethod
    def clear(self):
        pass  # pragma: no cover
